/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

/* Software by P.J.Sloetjes. */

"use strict";

const { classes: Cc, interfaces: Ci, utils: Cu } = Components,
	ID = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo).ID,
	FF = ID === "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}",
	PM = ID === "{8de7fcbb-c55c-4fbe-bfc5-fc555c87dbc4}",
	SM = ID === "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}",
	TB = ID === "{3550f703-e582-4d05-9a08-453d09bdfdc6}",
	BR = FF || PM || SM;
Cu.import("resource://gre/modules/Services.jsm", this);
FF && Cu.import("resource:///modules/CustomizableUI.jsm", this);
const file = Components.stack.filename,
	id = file.split(/\/([a-z]+)\@pjs\.nl/g)[1],
	UI = FF ? CustomizableUI || {} : {},
	XW = TB ? "mail:3pane" : "navigator:browser",
	RP = Services.io.getProtocolHandler("resource").
			QueryInterface(Ci.nsIResProtocolHandler);

const Util = {
	ns: Object.freeze({
		svg: "http://www.w3.org/2000/svg",
		xhtml: "http://www.w3.org/1999/xhtml",
		xul: "http://www.mozilla.org/keymaster/" +
				"gatekeeper/there.is.only.xul" }),
	elm: (doc, eid) =>
		doc && doc.getElementById(eid.replace("$", id)),
	on: (elm, add, eva, cap) => {
		if (elm)
			for (let evt in eva)
				add ? elm.addEventListener(evt, eva[evt], cap) :
					elm.removeEventListener(evt, eva[evt], cap);
	},
	nel: (elm, typ, eid, cls, att) => { const
		edo = elm.ownerDocument,
		doc = edo || elm,
		tys = typ.split(":"),
		eln = tys[1] ? doc.createElementNS(
				Util.ns[tys[0]] || Util.ns.xhtml, tys[1]) :
			doc.createElement(typ);
		eid && (eln.id = eid.replace(/\$/g, id));
		cls && (eln.className = cls.replace(/\$/g, id));
		for (let atr in att)
			eln.setAttribute(atr, att[atr]);
		edo && (elm.appendChild(eln));
		return eln
	},
	prevent: (evt) => {
		evt.preventDefault && evt.preventDefault();
		evt.stopPropagation && evt.stopPropagation();
	},
	log: (msg) => {
		if (msg === undefined) {
			msg = ""; 
			let sta = Components.stack;
			while (sta = sta.caller)
				msg = (sta.name || "") + "^" + msg
		}
		Services.console.logStringMessage(msg);
	}
};//Util

const Main = (() => { const
	state = Object.seal({ version: "" }),
	start = (ext) => {
		RP.setSubstitution(id, ext.resourceURI);
		Services.scriptloader.loadSubScript(
				"resource://" + id + "/" + id + ".js");
		Options.load();
		state.version = ext.version.
				match(/\d\.\d(.\d)?/)[0];
	},
	init = () => {
		Core.init && Core.init();
		Buttons.init();
		Windows.init();
		Frames.init();
	},
	stop = (ext, rsn) => {
		Options.save();
		if (rsn === APP_SHUTDOWN)
			return;
		Timers.exit();
		Buttons.exit();
		Windows.exit();
		Core.exit && Core.exit();
		Frames.exit();
		//RP.setSubstitution(id, null);
	},
	wipe = (ext, rsn) =>
		rsn === ADDON_UNINSTALL && Options.wipe();
	return Object.freeze({
		start, init, stop, wipe, 
		version: () => state.version
})})();//Main

var install = (ext, rsn) => {},
	startup = Main.start,
	shutdown = Main.stop,
	uninstall = Main.wipe;

const Options = (() => { const
	state = Object.create(null),
	map = new Map(),
	get = (opt) => 
		state[opt],
	set = (opt, val) =>
		(typeof val === typeof state[opt]) &&
				(state[opt] = val),
	def = (opt) =>
		map.get(opt),
	all = () =>
		map.keys(),
	com = { "pco": 1, "psz": 1, "pzm": 1, "tbb": 1 },
	init = (obj) => {
		for (let opt in obj)
			map.has(opt) || map.set(opt, obj[opt]);
	},
	pid = "extensions." + id + ".options",
	svc = Cc["@mozilla.org/content-pref/service;1"].
			getService(Ci.nsIContentPrefService2),
	load = () =>
		svc.getGlobal(pid, null, { 
				handleResult: onload,
				handleCompletion: onload, 
				handleError: () => {} }),				
	onload = (obj) => {
		if (state.pco !== undefined)
			return;
		map.has("pco") || map.set("pco", 0);
		map.has("psz") || map.set("psz", 2 * Panels.siz);
		map.has("pzm") || map.set("pzm", Panels.siz);
		map.has("tbb") || map.set("tbb", 0);
		//map.has("lan") || map.set("lan", "");
		for (let [opt, val] of map)
			state[opt] = val;
		Object.seal(state);
		obj = obj && JSON.parse(obj.value);
		if (typeof obj === "object") {
			for (let [opt, val] of map)
				if ((typeof obj[opt]) === (typeof val))
					state[opt] = obj[opt];
		} else if (Core.migratePrefs)
			Core.migratePrefs();
		$.load();
		try { Services.prefs.deleteBranch(
				"extensions." + id + "@pjs.nl.") }
		catch (err) {}
	},
	save = () =>
		state.pco !== undefined && svc.setGlobal(pid, 
				JSON.stringify(state), null),
	wipe = () => {
		svc.removeGlobal(pid, null);
		svc.removeByName("extensions." + id + ".theme", null);
		svc.removeByName("extensions." + id + ".design", null);
	};
	return Object.freeze({
		get, set, def, all, init, load, save, wipe, com
})})();//Options

const $ = (() => { const
	state = {},
	$ = (nam) => state[nam] || "",
	def = (jso) => { const
		lan = Cc["@mozilla.org/chrome/chrome-registry;1"].
				getService(Ci.nsIXULChromeRegistry).
				getSelectedLocale("global"),
		lal = lan.substring(0, 2);
		return jso && jso[lan] ? lan : 
				jso && jso[lal] ? lal : "en-US";
	},	
	load = () => { const
		xob = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].
				createInstance(Ci.nsIXMLHttpRequest);
		xob.open("GET", "resource://" + id + "/locales.json", true);
		xob.responseType = "json";
		xob.onreadystatechange = () =>
			xob.readyState === 4 && xob.status === 200 &&
					init(xob.response);
		xob.send();
	},
	init = (jso) => { const
		//lan = Options.get("lan") || def(loc);
		lan = def(jso), str = jso[lan] || {}, ref = jso["en-US"];
		//Options.set("lan", lan);
		for (let key in ref)
			state[key] = str[key] || ref[key];
		Main.init();
	};
	return Object.freeze({
		$, load
})})();//$

const Timers = (() => { const
	state = new Map(),
	set = (key, cbf, tms, typ) => {
		state.has(key) || state.set(key,
				Cc["@mozilla.org/timer;1"].
						createInstance(Ci.nsITimer));
		cbf && state.get(key).init(cbf,
				tms >= 0 ? tms : 0, 
				typ >= 0 ? typ : 0);
	},
	reset = (key) =>
		state.has(key) && state.get(key).cancel(),
	exit = () => {
		for (let [key, value] of state)
			reset(key);
		state.clear();
	};
	return Object.freeze({
		set, reset, exit
})})();//Timers

const Frames = (() => { const
	rand = () => Math.random().toString().replace(/\./g, ""),
	rid = rand(),
	path = "resource://" + id + "/bootframe.jsm?",
	gMsg = Cc["@mozilla.org/globalmessagemanager;1"].
			getService(Ci.nsIMessageListenerManager),
	send = (brw, mod, fun, arg) => { const 
		bmm = brw && brw.messageManager;
		bmm && bmm.sendAsyncMessage(id, { mod, fun, arg });
		if (bmm && !brw[id] && !Core.listen) {
			bmm.addMessageListener(id, Core.receive);
			bmm.sendAsyncMessage(id, { fun: "state" });
		}
	},
	broadcast = (mod, fun, arg) => 
		gMsg.broadcastAsyncMessage(id, { mod, fun, arg }),
	init = () => {
		if (Core.listen) {
			gMsg.addMessageListener(id, Core.listen);
			gMsg.loadFrameScript(path + rid, true);
		}
	},
	load = (brw) => { const
		bmm = !Core.listen && brw && brw.messageManager;
		if (bmm && !brw[id] && (brw[id] = rand())) {
			bmm.loadFrameScript(path + brw[id], false);
			bmm.addMessageListener(id, Core.receive);
		}
	},
	unload = (dox) => { const
		wix = dox && dox.defaultView,
		tbs = wix && wix.gBrowser && wix.gBrowser.mTabs;
		for (let tab of (tbs || [])) { const
			brw = tab.linkedBrowser, 
			bmm = brw && brw.messageManager;
			bmm && bmm.removeMessageListener(id, Core.receive);
			brw && (brw[id] = "");
		}
	},
	exit = () => {
		if (Core.listen) {
			gMsg.removeDelayedFrameScript(path + rid);
			gMsg.removeMessageListener(id, Core.listen);
		}
		broadcast("", "exit");
		Windows.all(unload);
	};
	return Object.freeze({
		init, exit, load, send, broadcast
})})();//Frames

const Windows = (() => { const
	state = Object.seal({ list: new Set() }),
	init = () => {
		all(delay);
		UI && UI.addListener && UI.addListener(onArea);
		Services.ww.registerNotification(obs);
	},
	exit = () => {
		state.list.clear();
		UI && UI.removeListener && UI.removeListener(onArea);
		Services.ww.unregisterNotification(obs);
		all(install, false);
	},
	obs = { observe: (wix, msg) => { const
		win = wix.QueryInterface(Ci.nsIInterfaceRequestor).
				getInterface(Ci.nsIDOMWindow);
		if (msg === "domwindowopened")
			Util.on(win, true, { load: onLoad });
		else if (msg === "domwindowclosed" && 
				isDox(win.document))
			install(win.document, false);
	}},
	onArea = {
		onAreaNodeRegistered: (are, con) =>
			delay(con.ownerGlobal.document)
	},
	isDox = (dox) =>
		dox && (dox = dox.documentElement) &&
				(dox.getAttribute("windowtype") === XW),
	onLoad = (evt) => { const
		dox = evt.target;
		Util.on(dox.defaultView, false, { load: onLoad });
		isDox(dox) && delay(dox, 1000);
	},
	delay = (dox, tim = 1000) => {
		state.list.add(dox);
		Timers.set("retry", retry, tim);
	},	
	retry = () => {
		for (let dox of state.list)
			install(dox, true);
	},
	install = (dox, bool = true) => {
		let iok = Window.init ? !!Window.init(dox, bool) : true;
		Menus.init(dox, bool);
		iok = iok && (UI.createWidget ? true : Buttons.append(dox, bool));
		iok = iok && !!Panels.init(dox, bool);
		bool && !iok ? delay(dox, 2000) : state.list.delete(dox);
	},
	all = (...arg) => { const 
		brs = Services.wm.getEnumerator(XW), cbf = arg[0];
		let wix = null;
		if (typeof cbf === "function")
			while (wix = brs.getNext())
				(arg[0] = wix.QueryInterface(Ci.nsIDOMWindow).document) && 
						(!arg[1] || arg[1] !== arg[0]) &&
					cbf.apply(null, arg);
	};
	return Object.freeze({
		init, exit, delay, all
})})();//Windows

const Menus = (() => { const
	items = Object.freeze(
		FF || PM ? { "contentAreaContextMenu": [ "$mi0" ],
				"tabContextMenu": [ "$mi1" ],
				"menu_ToolsPopup": [ "$mi2" ] } :
		SM ?  { "contentAreaContextMenu": [ "$mi0" ],
				"toolbar-context-menu": [ "$mi1" ],
				"taskPopup": [ "$mi2" ] } :
		TB ? { "mailContext": [ "$mi0" ], 
				"tabContextMenu": [ "$mi1" ],
				"taskPopup": [ "$mi2" ] } : null),
	init = (dox, bool) => { const
		its = Menu.items || items;
		Util.on(dox.defaultView, bool, { popupshowing });
		Menu.init && Menu.init(dox, bool);
		if (!bool)
			for (let min in its)
				for (let iid of (its[min] || []))
					if (iid = Util.elm(dox, iid))
						iid.parentNode.removeChild(iid);
	},
	popupshowing = (evt) => { const
		its = Menu.items || items,
		men = evt.target, ita = its[men.id];
		if (ita && ita.length) { const
			dox = evt.view.document, gcm = evt.view.gContextMenu;
			for (let iid of ita)
				if (typeof iid === "string") 
					(Menu.show || showItem)(Util.elm(dox, iid) || 
							addItem(dox, men, iid), gcm);
		}
	},
	addItem = (dox, men, iid) => { const
		sep = Util.elm(dox, men.id + "_pjs") || 
				Util.nel(men, "menuseparator", men.id + "_pjs"),
		elm = Util.nel(dox, "menuitem", iid, 
				"menuitem-iconic $mi", { label: $.$("nam") + 
						($.$(iid.replace("$", "")) ||  "...") }),
		ins = (sep.ins || sep).nextSibling;
		ins ? men.insertBefore(elm, ins) : men.appendChild(elm);
		Util.on(elm, true, { click: Menu.fun || Panels.show });
		return (sep.ins = elm);
	},
	showItem = (elm, gcm) =>
		elm.hidden = !!gcm && (gcm.onLink || gcm.onMailtoLink ||
				gcm.onTextInput || gcm.onImage || 
				gcm.isContentSelected || gcm.onMathML);
	return Object.freeze({
		init, showItem
})})();//Menus

const Buttons = (() => { const
	init = () =>
		UI && UI.createWidget && UI.createWidget({ 
				id, type: "custom", onBuild: create,
				removable: true, defaultArea: UI.AREA_NAVBAR }),
	create = (dox) => { const
		btn = Util.nel(dox, "toolbarbutton", "$", "toolbarbutton-1",
				{ type: Button.type, label: $.$("nam"), 
						style: "list-style-image:none!important;",
						tooltiptext: $.$("btt") });
		Util.on(btn, true, { click, dblclick: click });
		Button.update && Button.update(dox, btn);
		return btn;
	},
	click = (evt) => { const
		elm = evt && evt.originalTarget,
		but = elm.getAttribute && 
				elm.getAttribute("anonid") === "button";
		but ? evt.button || (Button.click && Button.click(evt)) :
				evt.button ? Util.prevent(evt) : Panels.show(evt);
	},
	append = (dox, bool) => { const
		btn = Util.elm(dox, id), tbb = Options.get("tbb"),
		box = Util.elm(dox, TB ? "mail-toolbox" : "navigator-toolbox"),
		bat = Util.elm(dox, TB ? "mail-bar3" : "nav-bar"),
		bip = box && box.palette && box.palette.
				getElementsByAttribute("id", id)[0];
		if (bool && !(box && box.palette && bat))
			return false;
		if (bool) { const
			elb = btn || create(dox),
			bar = dox.querySelector(
					"[currentset^='" + id + ",']," +
					"[currentset*='," + id + ",']," +
					"[currentset$='," + id + "']");
			bip || box.palette.appendChild(elb);
			if (bar) { const
				set = bar.getAttribute("currentset").split(",");
				let ind = set.indexOf(id) + 1, elb = null;
				while (ind < set.length && 
						!(elb = Util.elm(dox, set[ind])))
					ind++;
				bar.insertItem(id, elb, null, false);
			} else if (bat && tbb)
				bat.insertItem(id, null, null, false);
		} else if (!bool && (btn || bip)) { const
			but = btn || bip, bpn = but.parentNode;
			bpn && bpn.removeChild(but);
		}
		Util.on(dox.defaultView, bool, { 
				"aftercustomization": customize });
		return true;
	},
	customize = (evt) => { const
		dox = evt.target.document,
		btn = Util.elm(dox, id);
		Options.set("tbb", !!(btn && 
				btn.parentNode.id !== "BrowserToolbarPalette"));
	},
	exit = () =>
		UI && UI.destroyWidget && UI.destroyWidget(id);
	return Object.freeze({
		init, exit, append
})})();//Buttons

const Panels = (() => { const
	O = Options.get,
	siz = 264,
	init = (dox, bool) => { const
		mps = Util.elm(dox, "mainPopupSet"),
		pnl = Util.elm(dox, "$panel") || 
				(bool && mps && create(dox, mps));
		pnl && Util.on(pnl, bool, { 
				popupshowing: onShowPanel });
		!bool && pnl && pnl.parentNode.removeChild(pnl);
		return !!pnl;
	},
	create = (dox, mps) => { const 
		pnl = Util.nel(dox, "panel", "$panel", "$pnl", { 
				width: (Options.get("pzm") + "px") || "264px", 
				height: (Options.get("psz") + "px") || "512px",
				backdrag: "true", ignorekeys: "true", noautohide: "true" });
		mps.appendChild(pnl);
		addSheet(pnl, "default", true);
		addSheet(pnl, id, true, onStyled);
		header(pnl);
		view(pnl, "$frame", "panel", Panel.load);
		view(pnl, "$msg", "message", addMessage);
		view(pnl, "$ftr", "footer", footer);
		setColor(null, pnl);
		Util.on(pnl, true, { popupshowing: initViews });
		return pnl;
	},
	view = (pnl, pid, nam, cbf) => { const
		frm = Util.nel(pnl, "iframe", pid, "$frm", {
				src: "data:text/html;charset=utf-8," + 
						"<html><body></body></html>",
				disablehistory: "true", scrolling: "no", 
				sandbox: "", tooltip: "aHTMLTooltip" });
		frm.nam = nam;
		frm.cbf = cbf;
		return frm;
	},
	addSheet = (elm, nam, rep, cbf) => { const
		xob = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].
				createInstance(Ci.nsIXMLHttpRequest);
		xob.open("GET", "resource://" + id + "/" + nam + ".css", true);
		xob.overrideMimeType("text/plain");
		xob.onreadystatechange = onSheet.bind({ elm, nam, rep, cbf });
		xob.send();
	},
	onSheet = function (evt) {
		let xob = evt.target;
		if (xob.readyState !== 4 || xob.status !== 200)
			return;
		Util.nel(this.elm, "xhtml:style", (this.nam === id ? "" : id) +
				this.nam + "style", "", { type: "text/css" }).
			textContent = !this.rep ? xob.responseText :
					xob.responseText.replace(/\$/g, id).
							replace(/!/g, "!important;");
		this.cbf && this.cbf(this.elm);
		xob = xob.onreadystatechange = this.cbf = null;
	},
	onStyled = (elm) => {
		if (elm && Window.update)
			Window.update(elm.ownerDocument);
	},
	initViews = (evt) => { const
		pnl = evt.target;
		Util.on(pnl, false, { popupshowing: initViews });
		for (let frm of pnl.querySelectorAll("iframe")) { const
			doc = frm && frm.contentDocument, 
			bod = doc && doc.body;
			if (bod && !bod.childElementCount &&
					typeof frm.cbf === "function") {
				bod.style.display = "none";
				frm.cbf(bod);
			}
			if (bod && typeof frm.nam === "string") {
				addSheet(doc.documentElement, 
						frm.nam || "", false);
				frm.nam = "";
			}
		}
		size(pnl, Options.get("pzm"), Options.get("psz"));
		setColor(null, pnl);
	},
	addMessage = (bod) => { const
		frm = bod.ownerDocument.defaultView.frameElement,
		dox = frm.ownerDocument;
		if (frm.msg)
			return confirm(dox, frm.msg, frm.msf, frm.mse, frm.mss);
		let num = 0;
		while ($.$("mg" + num))
			num++;
		num = ~~(Math.random() * num);
		const txt = frm.msg || $.$("mg" + num) || "Please donate!";
		confirm(dox, txt, showLink, null, "dev");
	},
	confirm = (dox, msg, msf, mse, mss) => { const
		frm = Util.elm(dox, "$msg"),
		doc = frm.contentDocument,
		bod = (doc || {}).body,
		ini = bod && doc.readyState === "complete",
		pnl = Util.elm(dox, "$panel");
		frm.msg = ini ? "" : msg;
		frm.msf = ini ? null : msf;
		frm.mse = ini ? null : mse;
		frm.mss = ini ? "" : mss;
		if (!ini || pnl.state !== "open")
			open(dox, pnl);
		if (!ini)
			return;
		while (bod.firstChild) {
			bod.firstChild.mse && bod.firstChild.mse();
			bod.removeChild(bod.firstChild);
		}
		const elm = Util.nel(bod, "div", "", mss || "exc",
					{ a: $.$("lnk") + "/developers" }),
			ele = Util.nel(bod, "div", "", "clo");
		elm.textContent = msg;
		elm.title = msg;
		elm.msf = msf;
		ele.mse = mse;
		frm.style.setProperty("display", "", "important");
		Util.on(elm, true, { click: onConfirm });
		Util.on(ele, true, { click: onConfirm });
	},
	onConfirm = (evt) => { const
		elm = evt.target, msf = elm.msf || elm.mse,
		bod = evt.view.document.body,
		frm = evt.view.frameElement;
		while (bod.lastChild)
			bod.removeChild(bod.lastChild);
		frm.style.setProperty("display", "none", "important");
		msf && msf(evt);
	},
	onShowPanel = (evt) => { const
		pnl = evt.target;
		if (pnl.id === id + "panel") { const 
			dox = evt.view.document;
			Windows.all(Panels.hide, dox);
			if (Panel.update) { const
				frm = Util.elm(dox, "$frame"),
				doc = frm && frm.contentDocument;
				doc && Panel.update(dox, doc);
				pnl.moveTo(pnl.boxObject.screenX, 
						pnl.boxObject.screenY + 1);
			}
		}
	},
	show = (evt) => { const
		dox = evt.view.document,
		pnl = Util.elm(dox, "$panel");
		Windows.all(hide, dox);
		!pnl ? Windows.delay(dox, 1) : pnl.state === "closed" ? 
				open(dox, pnl) : pnl.hidePopup();
	},
	open = (dox, pnl) =>
		(pnl || Util.elm(dox, "$panel")).openPopup(Util.elm(dox, "$") ||
				Util.elm(dox, "PanelUI-menu-button"), "after_start"),
	hide = (dox) => { const
		pnl = Util.elm(dox, "$panel");
		pnl && pnl.hidePopup();
	},
	close = (evt) => { const
		pnl = evt.target.parentNode.parentNode.parentNode;
		pnl.hidePopup && pnl.hidePopup();
	},
	header = (pnl, str) => { const
		hbx = Util.nel(pnl, "hbox", "", "$hdr"),
		hdr = Util.nel(hbx, "hbox", "", "");
		Util.nel(hdr, "hbox", "", "$i16");
		Util.nel(hdr, "label", "", "$ttl",
				{ value: str || $.$("nam") + " " + 
						Main.version(), crop: "end" });
		Util.on(Util.nel(hdr, "div", "", "$pin", {
				tooltiptext: $.$("clo") }), 
						true, { click: close });
	},
	footer = (bod) => { const
		eid = [ "colorific", "colortransform", "columnsreader", 
				"extensor", "nosmalltext", "processtoggle", 
				"perfectview", "tabtweaks", "monitormaster"],
		ext = [ ["co", "colorific-1", "Colorific"],
				["ct", "color-transform", "Color Transform"],
				["cr", "column-reader", "Column Reader"],
				["ex", "extensor", "Extensor"],
				["ns", "no-small-text", "No Small Text"],
				["pt", "process-toggle", "Process Toggle"],
				["pv", "perfect-view", "Perfect View"],
				["tt", "tab-tweaks", "Tab Tweaks"],
				["wm", "monitor-master", "Window Master"] ],
		ine = eid.indexOf(id), hom = ext[ine] || [],
		lnk = $.$("lnk") + "/developers";
		Util.nel(bod, "div", "", "dev", { a: lnk, title: 
				$.$("crb") + " Peter J. Sloetjes (pjs.nl@live.com)" });
		Util.nel(bod, "div", "", "hom", { a: hom[1], 
				title: $.$("ohp") + " " + hom[2] });
		Util.nel(bod, "div", "", "pay", { a: lnk, title: $.$("dpp") });
		ext.splice(ine, 1);
		for (let cnt = 0; cnt < 3; cnt++) { const
			ind = ~~(Math.random() * ext.length), lin = ext[ind];
			Util.nel(bod, "div", "a", lin[0], { a: lin[1], 
					title: $.$("ohp") + " " + lin[2] });
			ext.splice(ind, 1);
		}
		Util.on(bod, true, { click: showLink });
		Util.on(Util.nel(bod, "div", "", "pco", {
				title: $.$("pco") }), true, { click: setColor, wheel: setColor });
		Util.on(Util.nel(bod, "div", "", "psz", {
				title: $.$("psz") }), true, { 
				mousedown: resize, dblclick: sizing, mouseup: resized });
		//Util.on(ico(hdr, "loc"), true, { click: setLocale });
	},
	showLink = (evt) => { const
		lnk = evt.target.getAttribute("a"),
		url = "https://addons.mozilla.org/en-US/firefox/addon/" + lnk;
		if (TB && lnk)
			Cc['@mozilla.org/messenger;1'].createInstance().
					QueryInterface(Ci.nsIMessenger).
				launchExternalURL(url);
		else if (lnk) { const
			gbr = evt.view.frameElement.ownerDocument.defaultView.gBrowser;
			gbr && (gbr.selectedTab = gbr.addTab(url));
		}
	},
	resize = (evt) => { const
		elm = evt.target;
		elm.setCapture(true);
		elm.clx = (evt.view.innerWidth - evt.clientX) || 0;
		elm.cly = evt.clientY || 0;
		Util.on(elm, true, { mousemove: sizing });
	},
	sizing = (evt) => { const
		pnl = evt.view.frameElement.parentNode, 
		elm = evt.target, box = pnl.boxObject,
		wth = evt.type === "dblclick" ? 1 : 
				evt.screenX - box.screenX + elm.clx,
		hgt = evt.type === "dblclick" ? Options.def("psz") : 
				evt.screenY - box.screenY + elm.cly;
		size(pnl, wth, hgt);
	},
	size = (pnl, wth, hgt) => { const
		bod = pnl.children[1].contentDocument.body,
		sup = bod.previousSibling, sdn = bod.nextSibling,
		pzm = +wth <= siz ? siz : +wth >= 2400 ? 2400 : +wth,
		psz = +hgt <= siz ? siz : +hgt >= 2400 ? 2400 : +hgt,
		sca = pzm / siz;
		Options.set("psz", psz);
		Options.set("pzm", pzm);
		pnl.setAttribute("width", pzm);
		pnl.setAttribute("height", psz);
		bod && (bod.style.transform = "scale(" + sca + ")");
		sup && (sup.style.transform = "scale(" + sca + ",1)");
		sdn && (sdn.style.transform = "scale(" + sca + ",-1)");
		updateSB(null, bod);
	},
	resized = (evt) => {
		Util.on(evt.target, false, { mousemove: sizing });
		evt.target.setCapture(false);
	},
	addScrolling = (bod) => { const
		doc = bod.ownerDocument, htm = doc.documentElement,
		elp = htm.insertBefore(Util.nel(bod, "div", "", "sup"), bod),
		eln = htm.appendChild(Util.nel(bod, "div", "", "sdn")); elp.i = -1; eln.i = 1;
		Util.on(elp, true, { click: clickSB }); Util.on(eln, true, { click: clickSB });
		Util.on(doc, true, { scroll: updateSB, MozScrolledAreaChanged: updateSB });
		Util.on(doc.defaultView, true, { resize: updateSB });
	},
	updateSB = (evt, elm) => { const 
		bod = elm || (evt.target.document || evt.view.document).body,
		elp = bod.previousSibling, eln = bod.nextSibling, top = bod.scrollTop,
		max = bod.scrollTopMax + 1, val = max / (bod.clientHeight + 1);
		elp.classList.toggle("sar", top > 1);
		eln.classList.toggle("sar", max > 2 && top < max - 3);
		elp.style.setProperty("--s", (25 * (    top / max)) + "%", "");
		eln.style.setProperty("--s", (25 * (1 - top / max)) + "%", "");
		elp.style.setProperty("--p", .15 + .1 * (val > 4 ? 4 : val), "");
		eln.style.setProperty("--p", .15 + .1 * (val > 4 ? 4 : val), "");
	},
	clickSB = (evt) =>
		evt.view.scrollByPages(evt.target.i, { behavior: "smooth" }),
	bgc = [ "255,255,255","204,221,255","0,17,34","17,17,17",
			"255,255,255","153,153,153","85,85,85","17,17,17",
			"238,255,221","238,221,255","34,17,51","0,0,0",
			"102,153,221","119,170,85","204,136,68","187,121,187",
			"34,136,187","85,153,51","170,119,51","170,85,136",
			"34,51,85","34,51,17","68,17,34","51,17,68",
			"17,17,51","0,34,0","51,0,17","34,0,34",
			"0,0,34","0,17,0","34,0,17","17,0,34"],
	fgc = [ "0,0,17","0,0,0","204,204,221","187,187,204",
			"34,34,34","0,0,0","238,238,238","204,204,204",
			"34,17,0","51,0,51","170,170,170","204,204,204",
			"0,0,0","0,0,0","0,0,0","0,0,0",
			"0,0,0","0,0,0","0,0,0","0,0,0",
			"255,255,255","255,255,255","255,255,255","255,255,255",
			"204,204,204","204,204,204","204,204,204","204,204,204",
			"187,187,187","187,187,187","187,187,187","187,187,187"],
	ctc = [ "0,17,85","36,51,102","121,136,187","68,102,153",
			"85,85,85","255,255,255","17,17,17","102,102,102",
			"119,51,17","102,34,85","170,119,170","170,119,0",
			"255,255,255","255,255,255","255,255,255","255,255,255",
			"255,255,255","255,255,255","255,255,255","255,255,255",
			"187,187,255","170,187,136","221,187,170","221,170,221",
			"187,153,51","187,153,51","170,136,51","170,136,34",
			"85,119,187","85,136,68","170,102,51","153,85,153"],
	setElmColor = (elm, num) => {
		elm && elm.style.setProperty("--bgc", 
				bgc[num % 32], "important");
		elm && elm.style.setProperty("--fgc", 
				fgc[num % 32], "important");
		elm && elm.style.setProperty("--ctc", 
				ctc[num % 32], "important");
	},
	setColor = (evt, pnl) => { const
		dox = (pnl || evt.view.frameElement).ownerDocument,
		pco = (+O("pco") + (evt ? 
				(evt.ctrlKey || evt.deltaY < 0) ? 31 : 1 : 0)) % 32;
		evt && Options.set("pco", pco);
		pnl = pnl || Util.elm(dox, "$panel");
		setElmColor(pnl, pco);
		for (let elm of pnl.childNodes) {
			setElmColor(elm, pco);
			if (elm = elm.contentDocument)
				setElmColor(elm.documentElement, pco);
		}
		Panel.setColor && Panel.setColor(dox);
	};
	return Object.freeze({
		init, show, hide, header, addScrolling, siz, view, confirm
})})();//Panels
