/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

/* Software by P.J.Sloetjes. */

"use strict";

const EXPORTED_SYMBOLS = ["colorizer"];

const colorizer = (() => { const
	cta = (d, l, t) => { const
		t0 = t[0]|0, t1 = t[1]|0,
		t2 = t[2]|0, t3 = t[3]|0,
		t4 = t[4]|0, t5 = t[5]|0,
		t6 = t[6]|0, t7 = t[7]|0;
		let r = 1, g = 1, b = 1, a = 1, 
			i = 1, n = (l - (l % 4))|0;
		if (n < 4 || n > (1<<26))
			return;
		while (n) {			
			n -= 4;

			r = (+d[n])|0;
			g = (+d[n + 1])|0;
			b = (+d[n + 2])|0;

			a = (r > g ? r : g)|0;
			i = (r < g ? r : g)|0;
			a = (a > b ? a : b)|0;
			i = (i < b ? i : b)|0;

			if (a !== i) {
				r = (r === a ? ((((g - b) << 20) 
						/ (6 * (a - i))) >> 8) + 
						(g < b ? 4096 : 0) :
					g === a ? ((((b - r) << 20)
						/ (6 * (a - i))) >> 8) + 1365 :
							  ((((r - g) << 20)
						/ (6 * (a - i))) >> 8) + 2731) & 16383;
				g = a + i + 1;
				g = (((a - i) << 12) / 
						(g < 256 ? g : 512 - g))|0;
			} else
				r = g = 0;

			r = ((t0 * r) >> 12) + t1;
			g = ((t2 * g) >> 12) + t3;
			b = ((t4 * (a + i + 1)) >> 9) + t5;

			g = g > 0 ? g : 0;
			a = r;
			r = b < 8192 ? b + ((b * g) >> 14) : 
					b + g - ((b * g) >> 14);
			g = b + b - r;
			b = 6 * (r - g);

			i = (a + 21845) & 16383;
			i = i < 2731 ? g + ((b * i) >> 14) :
				i < 8192 ? r : i > 10923 ? g :
				g + ((b * (10923 - i)) >> 14);
			d[n + 0] = i >> 6;

			i = (a + 16384) & 16383;
			i = i < 2731 ? g + ((b * i) >> 14) :
				i < 8192 ? r : i > 10923 ? g :
				g + ((b * (10923 - i)) >> 14);
			d[n + 1] = i >> 6;

			i = (a + 10923) & 16383;
			i = i < 2731 ? g + ((b * i) >> 14) :
				i < 8192 ? r : i > 10923 ? g :
				g + ((b * (10923 - i)) >> 14);
			d[n + 2] = i >> 6;
			
			d[n + 3] = (((d[n + 3] * t6) >> 8) + t7) >> 6;
		}
	},
	cth = (d, l, t) => { const
		t0 = t[0]|0, t1 = t[1]|0,
		t2 = t[2]|0, t3 = t[3]|0,
		t4 = t[4]|0, t5 = t[5]|0,
		t6 = t[6]|0, t7 = t[7]|0;
		let r = 1, g = 1, b = 1, a = 1, 
			i = 1, n = (l - (l % 4))|0;
		if (n < 4 || n > (1<<26))
			return;
		while (n) {
			n -= 4;

			r = (+d[n])|0;
			g = (+d[n + 1])|0;
			b = (+d[n + 2])|0;

			a = (r > g ? r : g)|0;
			i = (r < g ? r : g)|0;
			a = (a > b ? a : b)|0;
			i = (i < b ? i : b)|0;

			if (a !== i) {
				g = a + i + 1;
				g = (((a - i) << 12) / 
						(g < 256 ? g : 512 - g))|0;
				g = ((t2 * g) >> 12) + t3;
			} else
				g = t3;

			b = ((t4 * (a + i + 1)) >> 9) + t5;
			g = g > 0 ? g : 0;

			a = t1;
			r = b < 8192 ? b + ((b * g) >> 14) : 
					b + g - ((b * g) >> 14);
			g = b + b - r;
			b = 6 * (r - g);

			i = (a + 21845) & 16383;
			i = i < 2731 ? g + ((b * i) >> 14) :
				i < 8192 ? r : i > 10923 ? g :
				g + ((b * (10923 - i)) >> 14);
			d[n + 0] = i >> 6;

			i = (a + 16384) & 16383;
			i = i < 2731 ? g + ((b * i) >> 14) :
				i < 8192 ? r : i > 10923 ? g :
				g + ((b * (10923 - i)) >> 14);
			d[n + 1] = i >> 6;

			i = (a + 10923) & 16383;
			i = i < 2731 ? g + ((b * i) >> 14) :
				i < 8192 ? r : i > 10923 ? g :
				g + ((b * (10923 - i)) >> 14);
			d[n + 2] = i >> 6;
			
			d[n + 3] = (((d[n + 3] * t6) >> 8) + t7) >> 6;
		}
	},
	state = Object.seal({ rcn: null, rcv: new Map(), 
			rxn: null, rxv: new Map(), aad: false, ctp: null,
			CSSR: null, DOMU: null, IOSV: null, CPSV: null,
			ts: 0, te: 0, ti: 0, msg: "Colorizer" }),
	context = Object.seal({ rid: "", send: null, stop: null, log: null,
			Cc: null, Ci: null, Cu: null, Sv: null  }),
	file = Components.stack.filename,
	id = file.split(/\/([a-z]+)/g)[1],
	_id = "pjs_ct_" + file.split(/\?/g)[1],
	id_init = _id + "_init",
	id_res = _id + "_res",
	id_rid = _id + "_rid",
	id_cnv = _id + "_cnv",
	id_trf = _id + "_trf",
	id_cpt = _id + "_cpt",
	id_cpr = _id + "_cpr",
	id_tmp = _id + "_tmp",
	id_pref = "extensions." + id + ".theme",
	rgb_find = /rgb[^\)]*/g,
	var_find = /var\([^(\,|\)]*/g,
	sel_main = /(::?[^\,]+)|(\[[^\,]*\])/g,
	props = new Set([
		"background-color", "box-shadow",
		"background-image", "text-shadow",
		"border-top-color", "border-bottom-color",
		"border-left-color", "border-right-color",
		"border-image-source", "outline-color",
		"list-style-image", "color",
		"text-decoration-color" ]),
	names = //transparent|
"black|whitesmoke|antiquewhite|navajowhite|\
floralwhite|ghostwhite|white|lightslategrey|lightslategray|\
darkslategrey|darkslategray|slategrey|slategray|lightgrey|\
lightgray|darkgrey|darkgray|dimgrey|dimgray|graytext|grey|\
gray|silver|snow|mediumslateblue|lightsteelblue|cornflowerblue|\
darkslateblue|lightskyblue|midnightblue|deepskyblue|powderblue|\
dodgerblue|mediumblue|royalblue|cadetblue|slateblue|steelblue|\
lightblue|aliceblue|darkblue|skyblue|navy|blueviolet|blue|\
lightcyan|darkcyan|cyan|azure|mediumvioletred|palevioletred|\
mediumpurple|mediumorchid|darkmagenta|darkorchid|darkviolet|\
darkorange|orangered|indianred|lightpink|deeppink|hotpink|\
darkred|magenta|fuchsia|rebeccapurple|violet|orchid|orange|pink|\
red|indigo|crimson|lavenderblush|lavender|thistle|plum|teal|\
mediumturquoise|darkturquoise|paleturquoise|turquoise|purple|\
mediumaquamarine|aquamarine|aqua|mediumspringgreen|\
mediumseagreen|darkolivegreen|lightseagreen|darkseagreen|\
yellowgreen|lightgreen|springgreen|forestgreen|greenyellow|\
lawngreen|palegreen|limegreen|darkgreen|seagreen|green|lime|\
mintcream|honeydew|chartreuse|olivedrab|olive|ivory|beige|\
darkkhaki|khaki|lightgoldenrodyellow|darkgoldenrod|\
palegoldenrod|lightyellow|goldenrod|yellow|gold|lemonchiffon|\
cornsilk|oldlace|wheat|moccasin|papayawhip|blanchedalmond|tan|\
bisque|linen|peru|peachpuff|seashell|sienna|lightsalmon|\
darksalmon|salmon|burlywood|lightcoral|coral|tomato|mistyrose|\
gainsboro|chocolate|saddlebrown|sandybrown|rosybrown|firebrick|\
brown|maroon|-moz-fieldtext|-moz-field|-moz-dialogtext|-moz-dialog|\
-moz-comboboxtext|-moz-combobox|buttontext|buttonface|\
threedface|threedshadow|threedlightshadow|threeddarkshadow|\
-moz-hyperlinktext|-moz-activehyperlinktext|-moz-default-color|\
-moz-default-background-color|-moz-html-cellhighlighttext|\
-moz-html-cellhighlight|-moz-cellhighlighttext|-moz-cellhighlight|\
threedhighlight|highlighttext|highlight",
	vars =
"var(--in-content-page-color)|\
var(--in-content-page-background)|\
var(--in-content-text-color)|\
var(--in-content-selected-text)|\
var(--in-content-header-border-color)|\
var(--in-content-box-background)|\
var(--in-content-box-background-odd)|\
var(--in-content-box-background-hover)|\
var(--in-content-box-background-active)|\
var(--in-content-box-border-color)|\
var(--in-content-item-hover)|\
var(--in-content-item-selected)|\
var(--in-content-border-highlight)|\
var(--in-content-border-focus)|\
var(--in-content-border-color)|\
var(--in-content-category-text)|\
var(--in-content-category-border-focus)|\
var(--in-content-category-text-selected)|\
var(--in-content-category-background)|\
var(--in-content-category-background-hover)|\
var(--in-content-category-background-active)|\
var(--in-content-tab-color)|\
var(--in-content-link-color)|\
var(--in-content-link-color-hover)|\
var(--in-content-link-color-active)|\
var(--in-content-link-color-visited)|\
var(--in-content-primary-button-background)|\
var(--in-content-primary-button-background-hover)|\
var(--in-content-primary-button-background-active)|\
var(--in-content-table-border-dark-color)|\
var(--in-content-table-header-background)|\
var(--view-highlight-color)",
	vals = {
"var(--in-content-page-color)": "#424e5a",
"var(--in-content-page-background)": "#fbfbfb",
"var(--in-content-text-color)": "#333",
"var(--in-content-selected-text)": "#fff",
"var(--in-content-header-border-color)": "#c8c8c8",
"var(--in-content-box-background)": "#fff",
"var(--in-content-box-background-odd)": "#f3f6fa",
"var(--in-content-box-background-hover)": "#ebebeb",
"var(--in-content-box-background-active)": "#dadada",
"var(--in-content-box-border-color)": "#c1c1c1",
"var(--in-content-item-hover)": "rgba(0,149,221,0.25)",
"var(--in-content-item-selected)": "#0095dd",
"var(--in-content-border-highlight)": "#ff9500",
"var(--in-content-border-focus)": "#0095dd",
"var(--in-content-border-color)": "#c1c1c1",
"var(--in-content-category-text)": "#c1c1c1",
"var(--in-content-category-border-focus)": "#fff",
"var(--in-content-category-text-selected)": "#f2f2f2",
"var(--in-content-category-background)": "#424f5a",
"var(--in-content-category-background-hover)": "#5e6972",
"var(--in-content-category-background-active)": "#343f48",
"var(--in-content-tab-color)": "#424f5a",
"var(--in-content-link-color)": "#0095dd",
"var(--in-content-link-color-hover)": "#178ce5",
"var(--in-content-link-color-active)": "#ff9500",
"var(--in-content-link-color-visited)": "#551a8b",
"var(--in-content-primary-button-background)": "#0095dd",
"var(--in-content-primary-button-background-hover)": "#008acb",
"var(--in-content-primary-button-background-active)": "#006b9d",
"var(--in-content-table-border-dark-color)": "#d1d1d1",
"var(--in-content-table-header-background)": "#0095dd",
"var(--view-highlight-color)": "#f9f5e5" },
	start = (arg, win, dsh, ctx) => {
		//ctx.log("Start Colorizer " + ctx.rid);
		for (let key in context)
			context[key] = ctx[key];
		state.CSSR = context.Ci.nsIDOMCSSStyleRule;
		state.DOMU = context.Cc["@mozilla.org/inspector/dom-utils;1"].
				getService(context.Ci.inIDOMUtils);
		state.IOSV = context.Cc["@mozilla.org/network/io-service;1"].
				getService(context.Ci.nsIIOService);
		/*state.CPSV = context.Cc["@mozilla.org/content-pref/service;1"].
				getService(context.Ci.nsIContentPrefService2);*/
		state.msg = arg.msg;
		activate(null, arg);
	},
	stop = (win) => {
		for (let key in state)
			state[key] = undefined;
		//log("Stop Colorizer " + context.rid);
		for (let key in context)
			context[key] = undefined;
		return false;
	},
	log = (str) =>
		context.log && context.log(str),
	time = () =>
		log(state.te + " " + state.ti),
	selectedElement = (doc) => { const
		sel = doc.getElementById("selector" + context.rid),
		elm = sel && sel.target; 
		return elm && elm.ownerDocument ? elm : null;
	},
	activate = (doc, aad) =>
		typeof aad === "boolean" && (state.aad = aad),
	getTopDoc = (elm) => {
		let doc = (elm && elm.ownerDocument) || elm;
		while (doc && (elm = doc.defaultView.frameElement))
			doc = elm.ownerDocument;
		return doc;
	},
	isTopDoc = (doc) =>
		!!(doc && doc.defaultView && 
				!doc.defaultView.frameElement &&
				doc.documentURI),
	post = (doc, ctp) => {
		if (isTopDoc(doc) && ctp !== null) {
			state.ctp = ctp;
			if (ctp.par && (
					!doc.documentElement[id_init] ||
							selectedElement(doc))) {
				run(doc, state.ctp, true);
				state.ctp = null;
			} else
				doc.location = doc.documentURI.split("#")[0];
		}
	},
	DOMContentLoaded = (evt) => { const
		doc = evt.target, top = getTopDoc(doc);
		/*context.log("DOMContentLoaded: " + evt.target.pjs);
		doc.pjs = true;*/
		if (doc === top) {
			if (state.ctp !== null) {
				run(doc, state.ctp, true);
				state.ctp = null;
			} else if (state.aad)
				context.send(doc.defaultView, 
						"pref", doc.documentURI);
		} else if (top[id_cpt] && !doc[id_cpt]) { const
			ctp = top[id_cpt];
			if (ctp.opt && ctp.opt.icf)
				ctRun(doc, ctp, true);
		}
	},
	saveState = (arg, elm) => { const
		ela = elm[_id] || (elm[_id] = {});
		ela.opt = arg.opt;
		ela.sav = true;
		ela.par || (ela.par = {});
		for (let key in arg.par) { const
			trf = arg.par[key], tre = ela.par[key];
			if (tre && tre.length === 8 && trf.length === 8) {
				for (let ind = 0; ind < 8; ind += 2) {
					tre[ind] = trf[ind] * +tre[ind];
					tre[ind + 1] = trf[ind + 1] + 
							trf[ind] * +tre[ind + 1];
				}
			} else if (trf.length === 8)
				ela.par[key] = trf.slice();
		}
	},
	save = (doc, arg) => { const
		elm = selectedElement(doc) || doc.documentElement;
		elm && context.send(doc.defaultView, "save", 
				{ trf: elm[_id], act: arg });
	},
	run = (doc, arg, frm) =>
		isTopDoc(doc) && ctRun(doc, arg, !(frm === false)),
	ctRun = (doc, arg, frm) => { const 
		par = arg && arg.par, opt = arg && arg.opt;
		//log("run - start " + _id);
		if (!doc || !doc.documentElement || 
				!doc.getElementById || !par || !opt)
			return;
		//log("run - continue");
		const els = doc.defaultView.frameElement ? null :
				selectedElement(doc);
		els && (doc = els.ownerDocument);
		const bas = doc.querySelectorAll("base")[0];
		const bar = !!(bas && bas.getAttribute("href"));
		const abp = doc.documentURI.indexOf("about:") === 0;

		state.rcv.has("black") || defineNames(doc);

		const tra = {};
		let len = 0, val = 0;
		for (let key of props)
			if ((val = intParams(par[key])) && ++len)
				tra[key] = val;
		if (tra["color"] && !tra["text-decoration-color"])
			tra["text-decoration-color"] = tra["color"];
		const tri = intParams(par["image"]);
		const trc = intParams(par["canvas"]);
		if (!len && !tri && !trc)
			return;

		opt.icf = !!(opt.icf && frm);
		opt.mpx = opt.mpx || 1<<26;
		opt.itc = !els && !bar && !abp && opt.itc;

		//state.ts = new Date().getTime();
		doc[id_cpt] = Object.create(arg);
		ctElement(doc, els, tra, tri, trc, opt);
		//state.te = new Date().getTime() - state.ts;

		arg.sav && saveState(arg, els || doc.documentElement);
		//log("run - end");
	},
	intParams = (par) => {
		if (!par || par.length < 8 || (
				par[0] == 1 && par[1] == 0 && 
				par[2] == 1 && par[3] == 0 &&
				par[4] == 1 && par[5] == 0 &&
				par[6] == 1 && par[7] == 0))
			return null;
		const tra = new Int32Array(8);
		for (let ind = 0; ind < 8; ind++)
			tra[ind] = (16384 * par[ind]) |0;
		return tra;
	},
	defineNames = (doc) => { const 
		cnv = doc && doc.createElementNS(
				"http://www.w3.org/1999/xhtml", "canvas");
		if (!cnv || !cnv.getContext)
			return;
		const ctx = cnv.getContext("2d");
		cnv.width = cnv.height = 1;
		for (let nam of names.split("|")) {
			ctx.fillStyle = nam;
			ctx.fillRect(0, 0, 1, 1);
			state.rcv.set(nam, ctx.getImageData(0, 0, 1, 1).data);
		}
		state.rcn = new RegExp(names, "gi")
		for (let nam in vals) {
			ctx.fillStyle = vals[nam];
			ctx.fillRect(0, 0, 1, 1);
			state.rxv.set(nam, ctx.getImageData(0, 0, 1, 1).data);
			//log(nam + " = " + Array.join(state.rxv.get(nam), ","));
		}
		state.rxn = new RegExp(vars.replace(/\(/g, "\\(").
				replace(/\)/g, "\\)"), "gi");
	},
	elStyle = (elm, cpr, rls) => { const 
		sty = elm && elm.style;
		rls || (rls = elm && state.DOMU.getCSSStyleRules(elm));
		if (!sty || !rls)
			return {};
		let val = sty.getPropertyValue(cpr), 
			pri = sty.getPropertyPriority(cpr),
			cnt = rls.Count(), rul = null;
		while (cnt-- && !pri)
			(rul = rls.GetElementAt(cnt).QueryInterface(state.CSSR)) &&
					(rul = rul.style) && (!pri || !val) &&
					((pri = rul.getPropertyPriority(cpr)) || !val) &&
					(val = rul.getPropertyValue(cpr));
		return { val, pri }
	},
	replaceAtts = (doc) => { const
		att2css = Object.freeze({
			text:"color", color:"color", 
			bgcolor:"background-color", 
			background:"background-image",
			bordercolor:"border-color"}),
		att2elm = Object.freeze({
			bgcolor: new Set(["body","table","tbody","tr","td"]),
			text: new Set(["body"]), 
			color: new Set(["body","font","hr"]),
			background: new Set(["body","table","td"]),
			bordercolor: new Set(["table"])});
		let sty = {}, val = "";
		for (let elm of doc.querySelectorAll(
				"[text], [color], [bgcolor], [background], [bordercolor]"))
			for (let att in att2css)
				if ((val = elm.getAttribute(att)) &&
						att2elm[att].has(elm.localName) &&
						(!(sty = elStyle(elm, att2css[att])) || !sty.val ||
								elm.localName === "hr"))
					elm.style.setProperty(att2css[att], 
							att === "background" ? "url(" + val + ")" : val);
	},
	addSheet = (doc) => {
		let sht = doc.styleSheets, ind = sht.length;
		while (ind-- && sht[ind] && sht[ind].title !== _id);
		return ind >= 0 ? sht[ind] : (sht = doc.createElementNS(
				"http://www.w3.org/1999/xhtml", "style")) &&
						(ind = doc.documentElement) &&
						(sht = ind.insertBefore(sht, ind.firstChild)) &&
						(sht.title = _id) && sht.sheet;
	},
	ctInit = (doc, roi) => { const
		sht = addSheet(doc), bod = doc.body, del = doc.documentElement;
		if (!sht || !sht.insertRule)
			return;
		if (!doc.defaultView.frameElement) { const
			bgc = "background-color", tra = "transparent",
			dbc = elStyle(del, bgc).val, bbc = bod && elStyle(bod, bgc).val,
			dbn = !dbc || dbc === tra, dbi = elStyle(del, "background-image").val;
			if (dbn && (!bbc || bbc === tra || dbi)) { const
				mdb = "-moz-default-background-color";
				dbi ? del.style.setProperty(bgc, mdb, elStyle(del, bgc).pri) :
				bbc === tra ? bod.style.setProperty(bgc, mdb, elStyle(bod, bgc).pri) :
				sht.insertRule("body{" + bgc + ":" + mdb + "}", 0);
			}
		}
		sht.insertRule("html{color:-moz-default-color}", 0);
		sht.insertRule("::-moz-selection{color:white;background-color:black}", 0);
		sht.insertRule("a{color:" + ((bod && bod.getAttribute("link")) || 
						"-moz-hyperlinktext") + "}", 0);
		sht.insertRule("a:active{color:-moz-activehyperlinktext}", 0);
		sht.insertRule("input,select,textarea,fieldset,button,textbox{\
color:-moz-fieldtext;background-color:-moz-field;border:2px solid gray;border-radius:3px}", 0);
		sht.insertRule("input[disabled='true'],textarea[disabled='true'],\
textbox[disabled='true'],button[disabled='true'],checkbox[disabled='true']{color:graytext}", 0);
		sht.insertRule("page menulist,page .checkbox-check,page .radio-check,page tree,\
page richlistbox,page tabs{border:2px solid gray;border-radius:3px}", 0);
		sht.insertRule("page hbox.header,page treecols,page listheader{\
border-bottom:2px solid gray}", 0);
		sht.insertRule("table,td{border-color:lightgray}", 0);
		sht.insertRule("hr{border:1px solid lightgray;border-top-color:darkgray}", 0);
		roi && sht.insertRule("img,canvas{filter:brightness(70%)}", 0);
		roi && sht.insertRule("img:hover,canvas:hover{filter:none;\
transition:filter .3s ease-in-out!important;transition-delay: .3s!important}", 0);
	},
	ctSelected = (els) => {
		let elm = els, col = "";
		while (elm && elm.parentNode && 
				!(col = elStyle(elm, "background-color").val))
			elm = elm.parentNode;
		col && els.style.setProperty("background-color", col);
		elm = els;
		col = "";
		while (elm && elm.parentNode && 
				!(col = elStyle(elm, "color").val))
			elm = elm.parentNode;
		col && els.style.setProperty("color", col);
	},
	elements = (elm, tag, tree, sel) => { const 
		arr = [], els = tree && (tag || sel) ?
				elm.querySelectorAll(tag || sel) : {};
		if (els.length)
			for (let elm of els)
				arr.push(elm);
		if ((tag === "*" || elm.localName === tag) ||
				(!tag && sel && elm.mozMatchesSelector &&
						elm.mozMatchesSelector(sel)))
			arr.push(elm);
		return arr;
	},
	ctElement = (doc, els, tra, tri, trc, opt) => { const 
		ctRules = (sht) => {
			if (sht && sht.title && sht.title === _id)
				sht.disabled = false;
			else if (!sht || sht.disabled)
				return;
			let rls = [];
			try { rls = sht.cssRules }
			catch (err) { rls = [] }
			let ind = rls.length;
			const urb = ind > 0 ? state.IOSV.newURI(sht.href || 
					doc.baseURI, null, null) : null;
			while (ind--) { const 
				rul = rls.item(ind);
				rul.cssRules ? ctRules(rul) : rul.styleSheet ? 
						ctRules(rul.styleSheet) : ctProps(rul, urb);
			}
		},
		ctStyles = (elm) => { const 
			rls = state.DOMU.getCSSStyleRules(elm),
			sty = elm.style;
			if (rls && sty)
				for (let key in tra) { const
					spv = elStyle(elm, key, rls);
					if (spv && spv.val !== "")
						ctProp(sty, spv.val, key, null, spv.pri);
				}
		},
		ctProps = (elm, urb) => { const 
			sty = elm.style;
			if (sty)
				for (let key in tra) { const 
					val = sty.getPropertyValue(key);
					val !== "" && ctProp(sty, val, key, urb, undefined, 
							opt.bgr ? null : key[1] === 'a' && elm.selectorText);
				}
		},
		ctProp = (sty, str, cpr, urb, pri, sel) => { const
			ctStr = (str, trf) => { const
				ctS = (str) => { const 
					arr = str.substring(str.indexOf(
							"(") + 1).split(",");
					arr[3] = arr.length === 3 ? 255 :	
							(.5 + 255 * arr[3]) |0;
					cta(arr, 4, trf);
					arr[3] /= 255;
					return "rgba(" + arr.join(", ");
				},
				ctN = (nam) => { const
					arr = new Uint8ClampedArray(
							state.rcv.get(nam.toLowerCase()) ||
							state.rxv.get(nam.toLowerCase()));
					cta(arr, 4, trf);
					let opa = arr[3];
					return "rgba(" + arr[0] + ", " + arr[1] +
							", " + arr[2] + ", " + (opa / 255) + ")";
				};
				str = str.replace(rgb_find, ctS);
				return !abo ? str.replace(state.rcn, ctN) :
						str.replace(state.rxn, ctN).
							replace(state.rcn, ctN);
			},
			ctGrad = (str, trf) => {
				let ino = 0, ing = 0, stt = "";
				while (ing = str.indexOf("gradient(", ino) + 1) {
					stt += str.substring(ino, ing += 8);
					ino = str.indexOf("url(", ing);
					(ino < 0) && (ino = str.length);
					stt += ctStr(str.substring(ing, ino), trf || tra[cpr]);
				}
				return ctURL(ino ? stt + str.substr(ino) : str);
			},
			ctURL = (str) => {
				let ino = 0, inu = 0, set = new Set([]);
				while (inu = str.indexOf("url(", ino) + 1) {
					ino = str.indexOf(")", inu += 4) - 1;
					const uri = str.substring(inu, ino);
					if (uri.toLowerCase().indexOf(".svg") === -1)
						set.add(uri);
				}
				let ref = true, tim = false;
				if (typeof sel === "string") {
					try { ref = doc.querySelector(
							sel.replace(sel_main, "")) }
					catch (err) { ref = null }
					tim = ref && doc.querySelectorAll(
							sel.replace(sel_main, "")).length === 1 &&
							ref.getElementsByTagName("*").length < 2;
				}
				for (let uri of set) { const
					ura = (urb && uri.substring(0, 5) !== "data:" &&
							state.IOSV.newURI(uri, null, urb).spec) || uri,
					toc = opt.itc && cpr[1] === 'a';
					let cid = ref && ctImage(doc, ura, null, tra[cpr], toc ?
							{ ura, itc: true, cpr } : 
							{ ura, itc: false, cpr, sty }, tim ? 1<<14 : 0);
					cid && (str = str.replace('url("' + uri + '")', toc ?
							'-moz-element(#' + cid + ')' : 'url("' + cid + '")'));
				}
				return str;
			};
			sty.setProperty(cpr, cpr[cpr.length - 1] === 'e' ?
					str.length < 5 ? str : cpr[0] === 'b' ?
				ctGrad(str, tra[cpr]) : ctURL(str) : 
				ctStr(str, tra[cpr]), pri !== undefined ?
						pri : sty.getPropertyPriority(cpr));
		},
		ctImage = (doc, ura, img, trf, tgt, mpx) => { const
			map = doc[id_res] || (doc[id_res] = new Map()),
			uni = els ? Math.random().toString().replace(/\./g, "") : "",
			urt = ura + uni, rem = map.get(urt),
			res = rem || { img: (!uni && img) ? img : 
					new doc.defaultView.Image(), map: new Map() },
			mac = res.map, trs = arr2str(trf), ctm = mac.get(trs),
			ctc = ctm || { trf, cnv: null, c2b: false, bid: null, 
					mpx: mpx || 1<<26, fin: false, set: new Set(),
					cid: _id + "_" + map.size + "_" + mac.size };
			ctc.set.add(tgt);
			ctm || res.map.set(trs, ctc);
			img || (res.img[id_tmp] = true);
			if (!rem && typeof urt === "string") {
				map.set(urt, res);
				if (res.img.complete && res.img.src.length) {
					for (let [key, value] of res.map)
						ctCanvas(res.img, value);
				} else {
					res.img[id_res] = res;
					for (let evi of [ "load", "error", "abort" ])
						res.img.addEventListener(evi, ctFetch, false);
					if ((uni || !tgt.img) && tgt.ura)
						res.img.src = tgt.ura;
				}
			} else if (ctc.fin)
				ctTarget(tgt, ctc);
			return ctc.cid;
		},
		arr2str = (arr) => {
			let str = "";
			for (let val of arr)
				str += "," + val;
			return str;
		};
		const top = els || (doc && doc.documentElement),
			dur = doc && doc.baseURI && state.IOSV.newURI(
					doc.baseURI, null, null),
			abo = dur && dur.spec &&
					dur.spec.substring(0, 6) === "about:";
		if (!top)
			return;
		let ini = !top[id_init];
		ini && replaceAtts(doc);
		ini && ctInit(doc, opt.roi);
		els && /*opt.sic &&*/ ctSelected(els);
		top[id_init] = true;
		for (let cnv of elements(els || doc, "canvas", opt.ice))
			if (trc || tra[cnv[id_cpr]])
				ctData(cnv, cnv.width, cnv.height, tra[cnv[id_cpr]] || trc);
		if (els && ini)
			for (let elm of elements(els, "*", opt.ice))
				ctStyles(elm);
		else {
			els || Array.forEach(doc.styleSheets, ctRules);
			for (let elm of elements(els || doc, "", opt.ice, "[style]"))
				ctProps(elm, dur);
		}
		if (tri)
			for (let img of elements(els || doc, "img", opt.ice))
				img[id_tmp] || ctImage(doc, img.src, img, tri, 
						{ img, ura: img.src }, opt.mpx);
		for (let frm of elements(els || doc, "iframe", opt.icf)) { const
			dof = frm.contentDocument;
			if (dof && dof.readyState !== "loading" &&
					dof.readyState !== "uninitialized") {
				ctElement(dof, null, tra, tri, trc, opt);
				dof[id_cpt] = true;
			}
		}
	},
	ctContainer = (doc, cnv, cid) => {
		let con = doc[id_cnv];
		if (!con) {
			con = doc.createElementNS(
					"http://www.w3.org/1999/xhtml", "div");
			con.style.setProperty("display", "none", "important");
			con.id = id_cnv;
			doc.documentElement.appendChild(con);
			doc[id_cnv] = con;
		}
		if (con && !doc.getElementById(cid)) {
			con.appendChild(cnv);
			cnv.id = cnv.title = cid;
		}
	},
	ctFetch = (evt) => { const
		img = evt.target;
		for (let evi of [ "load", "error", "abort" ])
			img.removeEventListener(evi, ctFetch, false);
		if (evt.type === "load" && img[id_res]) {
			for (let [key, value] of img[id_res].map)
				ctCanvas(img, value);
			img[id_res] = null;
		}
	},
	ctCanvas = (img, ctc) => { const
		wth = img.naturalWidth, 
		hgt = img.naturalHeight,
		doc = img.ownerDocument;
		if (wth > 0 && hgt > 0 &&
				wth < 32768 && hgt < 32768 && (ctc.cnv ||
				(ctc.cnv = doc.createElementNS(
						"http://www.w3.org/1999/xhtml", "canvas")))) { const
			cnv = ctc.cnv;
			cnv.width = wth;
			cnv.height = hgt;
			ctData(cnv, wth, hgt, (wth * hgt) < ctc.mpx ? ctc.trf : null, img);
			ctc.fin = true;
			let itc = false;
			for (let tgt of ctc.set) {
				ctTarget(tgt, ctc);
				itc |= tgt.itc;
			}
			itc && ctContainer(doc, cnv, ctc.cid);
		}
		//state.ti = new Date().getTime() - state.ts;
		//log(state.ti + "(transform)");
	},
	ctData = (cnv, wth, hgt, trf, img) => { const
		ctx = cnv.getContext("2d");
		if (!ctx)
			return;
		ctx.globalCompositeOperation = "source-over";
		img && ctx.drawImage(img, 0, 0, wth, hgt);
		const imd = ctx.getImageData(0, 0, wth, hgt);
		trf && trf[0] ? cta(imd.data, imd.data.length, trf) : 
				trf && cth(imd.data, imd.data.length, trf);
		ctx.putImageData(imd, 0, 0);
	},
	ctTarget = (tgt, ctc) => {
		if (!tgt.itc) {
			if (!ctc.c2b) { const
				bur = ctc.cnv.ownerDocument.baseURI;
				ctc.c2b = true;
				if (bur.indexOf("about:") < 0 &&
						bur.indexOf("chrome:") < 0)
					ctc.cnv.toBlob(ctBlob.bind(ctc));
				else
					ctBlob(null, ctc, ctc.cnv.toDataURL("image/png"));
			} else if (ctc.bid) {
				if (tgt.sty)
					ctURLStyle(tgt.sty, tgt.cpr, ctc.cid, ctc.bid);
				else if (tgt.img && tgt.img.src !== ctc.bid)
					tgt.img.src = ctc.bid;
			}
		} else if (tgt.cpr)
			ctc.cnv[id_cpr] = tgt.cpr;
	},
	ctBlob = function (blb, ctc = this, dat = null) {
		ctc.bid = dat || ctc.cnv.ownerDocument.
				defaultView.URL.createObjectURL(blb);
		for (let tgt of ctc.set)
			if (!tgt.sty && tgt.img)
				tgt.img.src = ctc.bid;
			else if (tgt.sty)
				ctURLStyle(tgt.sty, tgt.cpr, ctc.cid, ctc.bid);
		/*this.img.onload = () => { // multiple references!
			uro.revokeObjectURL(bid);
			this.cnv.width = this.cnv.height = 0;
		}*/
		//state.ti = new Date().getTime() - state.ts;
		//log(state.ti + "(blob)");
	},
	ctURLStyle = (sty, cpr, cid, bid) =>
		sty.setProperty(cpr, sty.getPropertyValue(cpr).
				replace('url("' + cid + '")', 
						'url("' + bid + '")'),
				sty.getPropertyPriority(cpr));
	return Object.freeze({
		start, stop, run, time, save, 
		DOMContentLoaded, activate, post
})})();
